/* 
	Author 	: Thomas GIL (DotNetGuru.org)
	Date 	: 17 september 2006
	License : Public Domain
*/
using System;
using System.Diagnostics;
using System.Reflection;

namespace DotNetGuru.UnitTestFramework {
    public class TestRunner {
        const BindingFlags flags = BindingFlags.Instance | BindingFlags.Public | BindingFlags.DeclaredOnly;

        public static int Main(string[] args) {
            int result = 0;
            Assembly a = Assembly.GetExecutingAssembly();

            // Iterate on every "TestCase" child
            foreach (Type t in a.GetTypes()) {
                if (typeof(TestCase).IsAssignableFrom(t))
                    result += RunTest((TestCase)Activator.CreateInstance(t));
            }
            if (TestCase.NbFailures > 0) result = -1;
            return result;
        }

        private static int RunTest(TestCase instance) {
            int result = 0;
            Type t = instance.GetType();

            // Iterate on every public method of that class 
            foreach (MethodInfo mi in t.GetMethods(flags)) {
                if (mi.Name.StartsWith("Test")) {
                    bool expectedExceptionOccurred = false;
                    Type expectedExceptionType = null;

                    object[] attrs = mi.GetCustomAttributes(typeof(ExpectedException), false);
                    if (attrs.GetLength(0) > 0) {
                        ExpectedException expected = (ExpectedException)attrs[0];
                        expectedExceptionType = expected.ExceptionType;
                    }

                    try {
                        // Invoke the method
                        mi.Invoke(instance, null);
                    } catch (Exception e) {
                        if (e.InnerException != null) e = e.InnerException;

                        if (expectedExceptionType != null && e.GetType().FullName == expectedExceptionType.FullName) {
                            // This was expected. No problem
                            expectedExceptionOccurred = true;
                        } else {
                            Console.Error.WriteLine("[{0}.{1}]", t.Name, mi.Name);
                            Console.Error.WriteLine("Test execution problem:");
                            Console.Error.WriteLine(e);
                            result = -1;
                        }
                    }

                    if (expectedExceptionType != null && !expectedExceptionOccurred) {
                        TestCase.ExpectedException(new StackFrame(true), expectedExceptionType);
                        result = -1;
                    }
                }
            }
            return result;
        }
    }
}
